/*****************************************************************************

       Copyright   1995, 1996 Digital Equipment Corporation,
                       Maynard, Massachusetts.

                        All Rights Reserved

Permission to use, copy, modify, and distribute this software and its 
documentation for any purpose and without fee is hereby granted, provided  
that the copyright notice and this permission notice appear in all copies  
of software and supporting documentation, and that the name of Digital not  
be used in advertising or publicity pertaining to distribution of the software 
without specific, written prior permission. Digital grants this permission 
provided that you prominently mark, as not part of the original, any 
modifications made to this software or documentation.

Digital Equipment Corporation disclaims all warranties and/or guarantees  
with regard to this software, including all implied warranties of fitness for 
a particular purpose and merchantability, and makes no representations 
regarding the use of, or the results of the use of, the software and 
documentation in terms of correctness, accuracy, reliability, currentness or
otherwise; and you rely on the software, documentation and results solely at 
your own risk. 

******************************************************************************/

/*****************************************************************************
*
* The portions of this that were updated to support ReiserFS are copyright 1996, 1997, 1998, 1999,
* 2000 Hans Reiser, excepting those portions which were taken from reiserfs/journal.c which are
* copyright Chris Mason 1999, 2000, and those portions (taken from reiserfs/hashes.c) labeled as
* copyright Jeremy Fitzhardinge.
*
*****************************************************************************/

/*
 * Keyed 32-bit hash function using TEA in a Davis-Meyer function
 *   H0 = Key
 *   Hi = E Mi(Hi-1) + Hi-1
 *
 * (see Applied Cryptography, 2nd edition, p448).
 *
 * Jeremy Fitzhardinge <jeremy@zip.com.au> 1998
 * 
 */



/* This is a set of functions that provides minimal filesystem
 * functionality to the Linux bootstrapper.  All we can do is
 * open and read files... but that's all we need 8-)
 */

//#define __KERNEL__

#if 0
#define _LINUX_REISER_FS_SB
#define _LINUX_REISER_FS_H
#endif

#include <linux/kernel.h>
#include <linux/config.h>
#include <linux/string.h>
#include <linux/stat.h>
#include <linux/kdev_t.h>

#include <linux/locks.h>
#include <linux/init.h>
#include <linux/unistd.h>
#include <string.h>
#include <ctype.h>

//#include "inc/reiserfs_fs.h"
//#include "inc/reiserfs_fs_sb.h"

#include <linux/reiserfs_fs.h>
#include <linux/reiserfs_fs_sb.h>



#if 0
#define MM_DEBUG_KEY
#define DEBUG_REISERFS
#endif


struct m_reiserfs_super_block
{
  __u32 s_block_count;
  __u32 s_free_blocks;                  /* free blocks count    */
  __u32 s_root_block;           	/* root block number    */
  __u32 s_journal_block;           	/* journal block number    */
  __u32 s_journal_dev;           	/* journal device number  */
  __u32 s_orig_journal_size; 		
  __u32 s_journal_trans_max ;           /* max number of blocks in a transaction.  */
  __u32 s_journal_block_count ;         /* total size of the journal. can change over time  */
  __u32 s_journal_max_batch ;           /* max number of blocks to batch into a trans */
  __u32 s_journal_max_commit_age ;      /* in seconds, how old can an async commit be */
  __u32 s_journal_max_trans_age ;       /* in seconds, how old can a transaction be */
  __u16 s_blocksize;                   	/* block size           */
  __u16 s_oid_maxsize;			/* max size of object id array, see get_objectid() commentary  */
  __u16 s_oid_cursize;			/* current size of object id array */
  __u16 s_state;                       	/* valid or error       */
  char s_magic[12];                     /* reiserfs magic string indicates that file system is reiserfs */
  __u32 s_hash_function_code;		/* indicate, what hash function is being use to sort names in a directory*/
  __u16 s_tree_height;                  /* height of disk tree */
  __u16 s_bmap_nr;                      /* amount of bitmap blocks needed to address each block of file system */
  __u16 s_version;		/* I'd prefer it if this was a string,
                                   something like "3.6.4", and maybe
                                   16 bytes long mostly unused. We
                                   don't need to save bytes in the
                                   superblock. -Hans */
  __u16 s_reserved;
  char s_unused[128] ;			/* zero filled by mkreiserfs */
}__attribute__ ((__packed__)) ;


#if 0
struct m_reiserfs_sb_info
{
    struct buffer_head * s_sbh;                   /* Buffer containing the super block */
				/* both the comment and the choice of
                                   name are unclear for s_rs -Hans */
    struct reiserfs_super_block * s_rs;           /* Pointer to the super block in the buffer */
    struct buffer_head ** s_ap_bitmap;       /* array of buffers, holding block bitmap */
    struct reiserfs_journal *s_journal ;		/* pointer to journal information */
    unsigned short s_mount_state;                 /* reiserfs state (valid, invalid) */
  
				/* Comment? -Hans */
    void (*end_io_handler)(struct buffer_head *, int);
    hashf_t s_hash_function;	/* pointer to function which is used
                                   to sort names in directory. Set on
                                   mount */
    unsigned long s_mount_opt;	/* reiserfs's mount options are set
                                   here (currently - NOTAIL, NOLOG,
                                   REPLAYONLY) */

				/* Comment? -Hans */
    wait_queue_head_t s_wait;
				/* To be obsoleted soon by per buffer seals.. -Hans */
    atomic_t s_generation_counter; // increased by one every time the
    // tree gets re-balanced
    
    /* session statistics */
    int s_kmallocs;
    int s_disk_reads;
    int s_disk_writes;
    int s_fix_nodes;
    int s_do_balance;
    int s_unneeded_left_neighbor;
    int s_good_search_by_key_reada;
    int s_bmaps;
    int s_bmaps_without_search;
    int s_direct2indirect;
    int s_indirect2direct;
};

extern inline __u32 type2uniqueness (int type)
{
    switch (type)
		{
		case TYPE_STAT_DATA: return V1_SD_UNIQUENESS;
		case TYPE_INDIRECT: return V1_INDIRECT_UNIQUENESS;
		case TYPE_DIRECT: return V1_DIRECT_UNIQUENESS;
		case TYPE_DIRENTRY: return V1_DIRENTRY_UNIQUENESS;
		}
    return V1_ANY_UNIQUENESS;
}

#define is_direntry_cpu_key(key) (cpu_key_k_type (key) == TYPE_DIRENTRY)
#define is_direct_cpu_key(key)   (cpu_key_k_type (key) == TYPE_DIRECT)
#define is_indirect_cpu_key(key) (cpu_key_k_type (key) == TYPE_INDIRECT)
#define is_statdata_cpu_key(key) (cpu_key_k_type (key) == TYPE_STAT_DATA)

#define is_direntry_cpu_ih(ih) (is_direntry_cpu_key (&((ih)->ih_key)))
#define is_direct_cpu_ih(ih)   (is_direct_cpu_key (&((ih)->ih_key)))
#define is_indirect_cpu_ih(ih) (is_indirect_cpu_key (&((ih)->ih_key)))
#define is_statdata_cpu_ih(ih) (is_statdata_cpu_key (&((ih)->ih_key)))

#endif


#include <asm/uaccess.h>


#define REISERFS_SUPER_MAGIC 0x52654973
				/* used by file system utilities that look at the superblock, etc. */
#define REISERFS_SUPER_MAGIC_STRING "ReIsErFs"
#define REISER2FS_SUPER_MAGIC_STRING "ReIsEr2Fs"

static inline int is_reiserfs_magic_string (struct m_reiserfs_super_block * rs)
{
    return (!strncmp (rs->s_magic, REISERFS_SUPER_MAGIC_STRING, 
					  strlen ( REISERFS_SUPER_MAGIC_STRING)) ||
			!strncmp (rs->s_magic, REISER2FS_SUPER_MAGIC_STRING, 
					  strlen ( REISER2FS_SUPER_MAGIC_STRING)));
}

struct offset_m_v1
{
    __u32 k_offset;
    __u32 k_uniqueness;
} __attribute__ ((__packed__)) ;

struct offset_m_v2
{
    __u64 k_offset:60;
    __u64 k_type: 4;
} __attribute__ ((__packed__));


/* Key of an item determines its location in the S+tree, and
   is composed of 4 components */
struct m_key
{
    __u32 k_dir_id;    /* packing locality: by default parent
						  directory object id */
    __u32 k_objectid;  /* object identifier */
    union
	{
		struct offset_m_v1 k_offset_v1;
		struct offset_m_v2 k_offset_v2;
    }  u;
} ;


#define M_KEY_SIZE (sizeof(struct m_key))
#define SHORT_KEY_SIZE (sizeof (__u32) + sizeof (__u32))

#define REISERFS_FULL_KEY_LEN     4
#define REISERFS_SHORT_KEY_LEN    2
#define m_comp_keys(k1,k2)        m__comp_keys((k1), (k2), 1)
#define m_comp_short_keys(k1,k2)  m__comp_keys((k1), (k2), 0)


#define ITEM_VERSION_1 0
#define ITEM_VERSION_2 1

#define TYPE_STAT_DATA 0
#define TYPE_INDIRECT 1
#define TYPE_DIRECT 2
#define TYPE_DIRENTRY 3 
#define TYPE_ANY 15 // FIXME: comment is required

#define V1_SD_UNIQUENESS 0
#define V1_INDIRECT_UNIQUENESS 0xfffffffe
#define V1_DIRECT_UNIQUENESS 0xffffffff
#define V1_DIRENTRY_UNIQUENESS 500
#define V1_ANY_UNIQUENESS 555 // FIXME: comment is required


extern inline int m_le_key_version (struct m_key * key)
{
    int type;
    type = le16_to_cpu (key->u.k_offset_v2.k_type);
    if (type != TYPE_DIRECT && type != TYPE_INDIRECT && type != TYPE_DIRENTRY)
		return ITEM_VERSION_1;
	else
		return ITEM_VERSION_2;
}

extern inline int uniqueness2type (__u32 uniqueness)
{
    switch (uniqueness)
		{
		case V1_SD_UNIQUENESS: return TYPE_STAT_DATA;
		case V1_INDIRECT_UNIQUENESS: return TYPE_INDIRECT;
		case V1_DIRECT_UNIQUENESS: return TYPE_DIRECT;
		case V1_DIRENTRY_UNIQUENESS: return TYPE_DIRENTRY;
		}
    return TYPE_ANY;
}

#define M_COMP_KEYS m_comp_keys
#define M_COMP_SHORT_KEYS m_comp_short_keys

//#define GET_HASH_VALUE(offset) ((offset) & 0x7fffff80)
#define M_GET_HASH_VALUE(offset) ((offset) & 0x7fffff80LL)

#define JOURNAL_TRANS_HALF 1018   /* must be correct to keep the desc and commit structs at 4k */

#define REISERFS_ROOT_OBJECTID 2
#define REISERFS_ROOT_PARENT_OBJECTID 1

#define JOURNAL_BLOCK_COUNT 8192 /* number of blocks in the journal */

#define REISERFS_DISK_OFFSET_IN_BYTES (64 * 1024)

#define SD_OFFSET  0
#define SD_UNIQUENESS 0
#define DOT_OFFSET 1
#define DOT_DOT_OFFSET 2
/*
#define DIRENTRY_UNIQUENESS 500
*/

/* The result of the key compare */
#define FIRST_GREATER 1
#define SECOND_GREATER -1
#define KEYS_IDENTICAL 0
#define KEY_FOUND 1
#define KEY_NOT_FOUND 0

#define ITEM_VERSION_1 0
#define ITEM_VERSION_2 1

/*
#define I_K_KEY_IN_ITEM(p_s_ih, p_s_key, n_blocksize) \
          ( ! M_COMP_SHORT_KEYS(p_s_ih, p_s_key) &&  I_OFF_BYTE_IN_ITEM(p_s_ih, k_offset (p_s_key), n_blocksize) )
*/
struct m_item_head
{
  struct m_key ih_key; 	/* Everything in the tree is found by searching for it based on its key.*/
	union
	{
		__u16 ih_free_space_reserved; /* The free space in the last unformatted node of an indirect item if this
										 is an indirect item.  This equals 0xFFFF iff this is a direct item or
										 stat data item. Note that the key, not this field, is used to determine
										 the item type, and thus which field this union contains. */
		__u16 ih_entry_count;         /* Iff this is a directory item, this field equals the number of directory
										 entries in the directory item. */
	}  u;
	__u16 ih_item_len;           /* total size of the item body                  */
	__u16 ih_item_location;      /* an offset to the item body within the block  */
	__u16 ih_version;	             /* 0 for all old items, 2 for new
									ones. Highest bit is set by fsck
									temporary, cleaned after all done */
} ;

#define M_IH_SIZE (sizeof(struct m_item_head))

struct m_block_head
{
  __u16 blk_level;        /* Level of a block in the tree. */
  __u16 blk_nr_item;      /* Number of keys/items in a block. */
  __u16 blk_free_space;   /* Block free space in bytes. */
  __u16 blk_reserved;
	                      /* dump this in v4/planA */
  struct m_key  blk_right_delim_key; /* kept only for compatibility */
};

#define M_BLKH_SIZE (sizeof(struct m_block_head))


struct m1_stat_data
{
    __u16 sd_mode;	/* file type, permissions */
    __u16 sd_reserved;
    __u32 sd_nlink;	/* number of hard links */
    __u64 sd_size;	/* file size */
    __u32 sd_uid;		/* owner */
    __u32 sd_gid;		/* group */
    __u32 sd_atime;	/* time of last access */
    __u32 sd_mtime;	/* time file was last modified  */
    __u32 sd_ctime;	/* time inode (stat data) was last changed (except changes to sd_atime and sd_mtime) */
    __u32 sd_blocks;
    union
	{
	__u32 sd_rdev;
	} __attribute__ ((__packed__)) u;
} __attribute__ ((__packed__));

struct m_stat_data
{
    __u16 sd_mode;	/* file type, permissions */
    __u16 sd_reserved;
    __u32 sd_nlink;	/* number of hard links */
    __u64 sd_size;	/* file size */
    __u32 sd_uid;		/* owner */
    __u32 sd_gid;		/* group */
    __u32 sd_atime;	/* time of last access */
    __u32 sd_mtime;	/* time file was last modified  */
    __u32 sd_ctime;	/* time inode (stat data) was last changed (except changes to sd_atime and sd_mtime) */
    __u32 sd_blocks;
	__u32 sd_rdev;
}__attribute__ ((__packed__));
//
// this is 40 bytes long
//
#define M_SD_SIZE (sizeof(struct m_stat_data))

struct reiserfs_m_de_head
{
  __u32 deh_offset;		/* third component of the directory entry key */
  __u32 deh_dir_id;		/* objectid of the parent directory of the object, that is referenced
					   by directory entry */
  __u32 deh_objectid;		/* objectid of the object, that is referenced by directory entry */
  __u16 deh_location;		/* offset of name in the whole item */
  __u16 deh_state;		/* whether 1) entry contains stat data (for future), and 2) whether
					   entry is hidden (unlinked) */
} ;
#define M_DEH_SIZE sizeof(struct reiserfs_m_de_head)

struct m_disk_child {
  __u32       dc_block_number;              /* Disk child's block number. */
  __u16       dc_size;		            /* Disk child's used space.   */
  __u16       dc_reserved;
};
#define M_DC_SIZE (sizeof(struct m_disk_child))


/* first block written in a commit.  */
struct reiserfs_m_journal_desc {
  __u32 j_trans_id ;			/* id of commit */
  __u32 j_len ;			/* length of commit. len +1 is the commit block */
  __u32 j_mount_id ;				/* mount id of this trans*/
  __u32 j_realblock[JOURNAL_TRANS_HALF] ; /* real locations for each block */
  char j_magic[12] ;
} ;

/* last block written in a commit */
struct reiserfs_m_journal_commit {
  __u32 j_trans_id ;			/* must match j_trans_id from the desc block */
  __u32 j_len ;			/* ditto */
  __u32 j_realblock[JOURNAL_TRANS_HALF] ; /* real locations for each block */
  char j_digest[16] ;			/* md5 sum of all the blocks involved, including desc and commit. not used, kill it */
} ;

/* this header block gets written whenever a transaction is considered fully flushed, and is more recent than the
** last fully flushed transaction.  fully flushed means all the log blocks and all the real blocks are on disk,
** and this transaction does not need to be replayed.
*/
struct reiserfs_m_journal_header {
  __u32 j_last_flush_trans_id ;		/* id of last fully flushed transaction */
  __u32 j_first_unflushed_offset ;      /* offset in the log of where to start replay after a crash */
  __u32 j_mount_id ;
} ;


#define M_MAX_ITEM_LEN(block_size) (block_size - M_BLKH_SIZE - M_IH_SIZE)

#define MIN_ITEM_LEN 1

#include "fs.h"

#define REISERFS_BLOCK_SIZE 4096

struct reiserfs_m_inode
{
	struct m_key i_key ;
	int i_data_length;
	__u32 i_data [2];

//	__u32 i_data [REISERFS_N_BLOCKS];
//	M_MAX_ITEM_LEN(REISERFS_BLOCK_SIZE)/sizeof(__u32)];

	__u16 i_mode;	/* file type, permissions */
    __u32 i_uid;		/* owner */
    __u64 i_size;	/* file size */
    __u32 i_atime;	/* time of last access */
    __u32 i_ctime;	/* time inode (stat data) was last changed (except changes to sd_atime and sd_mtime) */
    __u32 i_mtime;	/* time file was last modified  */
    __u32 i_dtime;	/* time file was last modified  */
    __u32 i_gid;		/* group */
    __u32 i_nlink_count;	/* number of hard links */

    __u32 i_blocks;
	__u32 i_flags;	/* File flags */
	__u32 i_version;	/* File version (for NFS) */
};

struct milo_item
{
	struct m_key ih_key;
	union
	{
		__u16 ih_free_space; 
		__u16 ih_entry_count; 
	} u;
	__u16 ih_item_len;
	__u16 ih_item_location;
	__u16 ih_version;

	__u16 it_Last;
	__u16 it_Number;

	struct m_key ih_r_key;

	union
	{
		struct m1_stat_data sd1;
		struct m_stat_data sd;
		struct reiserfs_m_de_head deh[(M_MAX_ITEM_LEN(REISERFS_BLOCK_SIZE)-M_DEH_SIZE)/M_DEH_SIZE];
		__u32 item[M_MAX_ITEM_LEN(REISERFS_BLOCK_SIZE)/sizeof(__u32)];
		char txt[M_MAX_ITEM_LEN(REISERFS_BLOCK_SIZE)];
	} body;
} ;


struct reiserfs_journal_ptr
{
	__u32 r_BlockNr;
	__u32 j_BlockNr;
};

static struct reiserfs_journal_ptr *j_table=NULL;

static __u32 j_tab_count = 0;

static struct reiserfs_m_journal_desc * desc;
static struct reiserfs_m_journal_commit * commit;
static struct reiserfs_m_journal_header * jh ;

static __u32	j_start ;
static __u32	j_last_flush_trans_id ;
static __u32	j_trans_id ;
static __u32   j_mount_id;

#define MAX_OPEN_FILES		4

static int fd ;

static unsigned long ErrorNumber;

static struct milo_item n_item_buf;
static struct milo_item it_buf;
static struct milo_item it_open;

static struct m_item_head * ih;
static struct m_block_head * bh;
static struct m_reiserfs_super_block sb;

static struct m_key m_root_key = {REISERFS_ROOT_PARENT_OBJECTID, REISERFS_ROOT_OBJECTID, 0, 0};
static hashf_t p_hash_function;

#define VV_ITEMS(bh)	  	   ((__u32)((bh)->blk_nr_item))
#define VV_CHILD(bh,n_pos)     ((struct m_disk_child *)((char *)(bh)+M_BLKH_SIZE+VV_ITEMS(bh)*M_KEY_SIZE+M_DC_SIZE*(n_pos)))
#define VV_CHILD_NUM(bh,n_pos) ((__u32)VV_CHILD((bh),(n_pos))->dc_block_number)
#define VV_KEY(bh,n_pos)       ((struct m_key *)((char *)(bh)+M_BLKH_SIZE+M_KEY_SIZE*(n_pos)))
#define VV_IHEAD(bh,n_pos)     ((struct m_item_head *)((char *)(bh)+M_BLKH_SIZE+M_IH_SIZE*(n_pos)))
#define VV_ITEM(bh,n_pos)      ((void *)((char *)(bh)+(VV_IHEAD((bh),(n_pos)))->ih_item_location))
#define VV_ITEM_SIZE(bh,n_pos) ((unsigned short)((char *)(bh)+(VV_IHEAD((bh),(n_pos)))->ih_item_len))

#define VV_M_IHEAD(p_i,bh,n_pos) vv_m_ihead((p_i),(bh),(n_pos))

#ifdef DEBUG_REISERFS
#define m_debug printk
#else
#define m_debug(a...)
#endif

#define m_copy_key(to,from)        memcpy((to), (from), M_KEY_SIZE )
#define m1_copy_key(to,from) {\
if (m_le_key_version (from)== ITEM_VERSION_1)\
{\
	memcpy((to), (from), SHORT_KEY_SIZE ); \
((struct m_key *)(to))->u.k_offset_v2.k_offset=\
	((struct m_key *)(from))->u.k_offset_v1.k_offset;\
((struct m_key *)(to))->u.k_offset_v2.k_type=\
	uniqueness2type(((struct m_key *)(from))->u.k_offset_v1.k_uniqueness);\
}\
else memcpy((to), (from), M_KEY_SIZE );\
}

/*
#define m_copy_short_key(to,from)  memcpy((to), (from), SHORT_KEY_SIZE )
#define m_copy_item_head(to,from)  memcpy((to), (from), M_IH_SIZE)
*/

static int blocksize;
static long s_dev;
//atic char filename[256];

static char blkbuf[REISERFS_BLOCK_SIZE];
static char iblkbuf[REISERFS_BLOCK_SIZE];






static struct inode_table_entry
{
	struct reiserfs_m_inode inode;
	int inumber;
	int free;
	unsigned short old_mode;
} inode_table[MAX_OPEN_FILES];








static struct reiserfs_m_inode *reiserfs_m_iget( struct m_key * ino_k, struct milo_item * tmp_item);
static int reiserfs_m_bread(int fd, long blkno, char *buffer);
static void reiserfs_m_breadi(struct reiserfs_m_inode *ip, long blkno, char *buffer);
static int reiserfs_m_mount(long device, int quiet);
static int reiserfs_m_open(char *filename);
static void reiserfs_m_cat(struct reiserfs_m_inode *ip, char *buffer);
static void reiserfs_m_ls(char *dirname, char *devname);
static void reiserfs_m_close(int fd);
static long reiserfs_m_fdsize(int fd);






struct bootfs reiserfs =
{
	"reiserfs",
	0,
	0,

	reiserfs_m_mount,
	reiserfs_m_open,
	reiserfs_m_fdsize,
	reiserfs_m_bread,
	reiserfs_m_ls,
	reiserfs_m_close,
};









/*>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>*/

#define bread_desc(a,b,c) mm_bread((a),(b),(c),(char *)desc)
#define bread_commit(a,b,c) mm_bread((a),(b),(c),(char *)commit)
#define bread_jh(a,b,c) mm_bread((a),(b),(c),(char *)jh)


static inline int mm_bread(int dev, __u32 blkno, long b_size, char * buf)
{
	__u32 ret;
	ret = device_read(fd, buf, b_size,(long)( blkno*b_size));
	if (ret != b_size)
		{
			printk("\nBread(inline) error on read blkno %d=%d/%d/%d/%p",ret,dev,blkno, b_size,buf);
			return -1;
		}
	return 0;
}

static __u32 check_j_block(__u32 p_blkn)
{
	int i,l,k;
	__u32 ret;
	l=j_tab_count;
	k=0;
	i=(k+l)/2;
	ret =p_blkn;
	while (k<l)
		{
			if (p_blkn>(j_table+i)->r_BlockNr ) k=i+1;
			if (p_blkn<(j_table+i)->r_BlockNr ) l=i-1;
			if (p_blkn==(j_table+i)->r_BlockNr ) { l=k=i; ret=(j_table+i)->j_BlockNr;}
			i=(k+l)/2;
		}
	
	if ( p_blkn != ret ) m_debug("(%d:%d)",p_blkn,ret);
	return (ret);
}

/* compares description block with commit block.  returns 1 if they differ, 0 if they are the same */
static int journal_compare_desc_commit(struct reiserfs_m_journal_desc *p_desc, 
									   struct reiserfs_m_journal_commit *p_commit)
{
	if (p_commit->j_trans_id != p_desc->j_trans_id ||
		p_commit->j_len != p_desc->j_len ||
		p_commit->j_len > JOURNAL_TRANS_MAX || 
		p_commit->j_len <= 0 )
		{
			return 1 ;
		}
  return 0 ;
}

/* returns 0 if it did not find a description block  
** returns -1 if it found a corrupt commit block
** returns 1 if both desc and commit were valid 
*/
static int journal_transaction_is_valid(__u32 p_blknr,
										__u32 *oldest_invalid_trans_id,
										__u32 *newest_mount_id)
{

	struct m_reiserfs_super_block *rs = &sb ;

	__u32 offset ;

	if (desc->j_len > 0 && !memcmp(desc->j_magic, JOURNAL_DESC_MAGIC, 8))
		{
			if (oldest_invalid_trans_id &&
				*oldest_invalid_trans_id && 
				desc->j_trans_id > *oldest_invalid_trans_id)
				{
					return 0 ;
				}
			if (newest_mount_id && *newest_mount_id > desc->j_mount_id)
				{
					return -1 ;
				}
			offset = p_blknr - rs->s_journal_block ;
		
			/* ok, we have a journal description block, lets see if the 
			** transaction was valid 
			*/
			if (bread_commit(s_dev,
							  rs->s_journal_block + ((offset + desc->j_len + 1) % JOURNAL_BLOCK_COUNT), rs->s_blocksize) )
				return 0 ;
			m_debug("c");
			
			if (journal_compare_desc_commit(desc, commit))
				{
					if (oldest_invalid_trans_id)
						*oldest_invalid_trans_id = desc->j_trans_id ;
					return -1; 
				}
			return 1 ;
		}
	else
		{
			return 0 ;
		}
}




static int journal_read_transaction(__u32 cur_dblock,
									__u32 oldest_start, 
									__u32 oldest_trans_id,
									__u32 newest_mount_id)
{
	struct m_reiserfs_super_block *rs = &sb ;
	__u32 trans_id = 0 ;
	__u32 trans_offset ;
	int i;

	if (bread_desc(s_dev, cur_dblock, rs->s_blocksize) )
		return 1 ;

	m_debug("d");

	trans_offset = cur_dblock - rs->s_journal_block ;

	if (desc->j_trans_id < oldest_trans_id)
		{
			return 1 ;
		}
	if (desc->j_mount_id != newest_mount_id)
		{
			return 1 ;
		}
	if (bread_commit(s_dev, rs->s_journal_block + ((trans_offset + desc->j_len + 1) % JOURNAL_BLOCK_COUNT), 
					  rs->s_blocksize) )
			return 1 ;
	m_debug("c");
	if (journal_compare_desc_commit(desc, commit))
		{
			return 1; 
		}
	trans_id = desc->j_trans_id ;

	/* now we know we've got a good transaction, and it was inside the valid time ranges */

	m_debug("\nTransaction_id=%d len=%d mount=%d ",desc->j_trans_id,desc->j_len,desc->j_mount_id);
	
	/* get all the buffer heads */
	for(i = 0 ; i < desc->j_len ; i++)
		{
			(j_table+j_tab_count)->j_BlockNr =  rs->s_journal_block + (trans_offset + 1 + i) % JOURNAL_BLOCK_COUNT;
			(j_table+j_tab_count)->r_BlockNr =  ((i < JOURNAL_TRANS_HALF)?(desc->j_realblock[i]):(commit->j_realblock[i - JOURNAL_TRANS_HALF]));
				
			if ((j_table+j_tab_count)->r_BlockNr >= rs->s_journal_block &&
				(j_table+j_tab_count)->r_BlockNr < (rs->s_journal_block+JOURNAL_BLOCK_COUNT))
				{
					printk("clm-2026: REPLAY FAILURE fsck required! Trying to replay onto a log block\n") ;
					return -1 ;
				}
			j_tab_count++;
			m_debug("!%d:(%d/%d)",j_tab_count,(j_table+j_tab_count)->r_BlockNr,(j_table+j_tab_count)->j_BlockNr);
		}
	m_debug("*");
	cur_dblock = rs->s_journal_block + ((trans_offset + desc->j_len + 2) % JOURNAL_BLOCK_COUNT) ;


	/* init starting values for the first transaction, in case this is the last transaction to be replayed. */
	j_start = cur_dblock - rs->s_journal_block ;
	j_last_flush_trans_id = trans_id ;
	j_trans_id = trans_id + 1;
	return 0 ;
}


/*
** read and replay the log
** on a clean unmount, the journal header's next unflushed pointer will be to an invalid
** transaction.  This tests that before finding all the transactions in the log, whic makes normal mount times fast.
**
** After a crash, this starts with the next unflushed transaction, and replays until it finds one too old, or invalid.
**
** On exit, it sets things up so the first transaction will work correctly.
*/
static int m_journal_read()
{
	struct m_reiserfs_super_block *rs = &sb;
	__u32 last_flush_trans_id = 0 ;
	__u32 oldest_trans_id = 0;
	__u32 oldest_invalid_trans_id = 0 ;
	time_t start ;
	__u32 last_flush_start = 0;
	__u32 oldest_start = 0;
	__u32 cur_dblock = 0 ;
	__u32 newest_mount_id = 9 ;
	int valid_journal_header = 0 ;
	int replay_count = 0 ;
	int continue_replay = 1 ;
	int ret ;
	int tmp_i;
	
	cur_dblock = rs->s_journal_block ;
	start = CURRENT_TIME ;

	/* step 1, read in the journal header block.  Check the transaction is says is the first unflushed,
	** and if that transaction is not valid, replay is done
	*/


	if( !( jh=kmalloc( REISERFS_BLOCK_SIZE,0) ) )
		{
			printk("\n can't allocate jh");
			return -1;
		}
	if ( bread_jh(s_dev, rs->s_journal_block +	JOURNAL_BLOCK_COUNT,rs->s_blocksize) )
		{
			return -1;
		}
	m_debug("\nH");
	if (jh->j_first_unflushed_offset >= 0 &&
		jh->j_first_unflushed_offset < JOURNAL_BLOCK_COUNT &&
		jh->j_last_flush_trans_id > 0)
		{
			last_flush_start = rs->s_journal_block + jh->j_first_unflushed_offset ;
			last_flush_trans_id = jh->j_last_flush_trans_id ;
			valid_journal_header = 1 ;
			
			/* now, we try to read the first unflushed offset.  If it is not valid, there is nothing more we can do,
			** and it makes no sense to read through the whole log.
			*/

			desc   = kmalloc( REISERFS_BLOCK_SIZE,0);
			commit = kmalloc( REISERFS_BLOCK_SIZE,0);

			if (!desc || !commit)
				{
					printk("\ncan't allocate desk or commit. reboot ");
					if (desc) kfree(desc);
					if (commit) kfree(commit);
					return -1;
				}

			bread_desc(s_dev, last_flush_start, rs->s_blocksize) ;
			m_debug("d");
			ret = journal_transaction_is_valid(last_flush_start,NULL, NULL) ;

			if (!ret)
				{
					continue_replay = 0 ;
				}
		}
	
	/* ok, there are transactions that need to be replayed.  
	** start with the first log block, find all the valid transactions, 
	** and pick out the oldest.
	*/
	if (continue_replay) printk("\nNeed Journal replay. Please wait.");
	tmp_i=0;
	while(continue_replay  &&   cur_dblock < (rs->s_journal_block + JOURNAL_BLOCK_COUNT) )
		{
			if ( ! ( (tmp_i++) % 256 ))printk(".");
			bread_desc(s_dev, cur_dblock, rs->s_blocksize) ;
			if ((ret = journal_transaction_is_valid(cur_dblock,&oldest_invalid_trans_id, 
													&newest_mount_id)) == 1)
				{
//					m_debug("d=%d(%d)",cur_dblock,desc->j_len);
					if ( oldest_start == 0 || newest_mount_id < desc->j_mount_id)
						{
							oldest_trans_id = desc->j_trans_id ;
							oldest_start = cur_dblock ;
							newest_mount_id = desc->j_mount_id ;
						}
					else
						{
							if (oldest_trans_id > desc->j_trans_id)  /* one just read was older */
								{
									oldest_trans_id = desc->j_trans_id ;
									oldest_start = cur_dblock ;
								}
						}
					cur_dblock += desc->j_len + 2 ;
				} 
			else
				{
					cur_dblock++ ;
				}
		}
//	m_debug("\n->%d<-",cur_dblock);
	/* step three, starting at the oldest transaction, replay */
	if (last_flush_start > 0)
		{
			oldest_start = last_flush_start ;
			oldest_trans_id = last_flush_trans_id ;
		} 
	cur_dblock = oldest_start ;

	if (oldest_trans_id)
		{
			m_debug("\nclm-2035: Starting replay from offset %lu, trans_id %lu\n",
				   cur_dblock - rs->s_journal_block, oldest_trans_id) ;
		}

	replay_count = 0 ;
	if (!j_table)
		j_table=( struct reiserfs_journal_ptr *)kmalloc(REISERFS_BLOCK_SIZE*8,0);
	if (!j_table)
		{
			printk("\ncan't allocate j_table");
			return -1;
		}
	
	j_tab_count=0;

	m_debug("%d/%d/%d/",oldest_start, oldest_trans_id, newest_mount_id);
	while(continue_replay && oldest_trans_id > 0)
		{
			printk(".");
			ret = journal_read_transaction(cur_dblock, oldest_start, oldest_trans_id, newest_mount_id) ;
			if (ret < 0)
				{
					continue_replay =0;
				}
			else
				if (ret != 0)        /* ==1 */
					{
						break ;
					}
			cur_dblock = rs->s_journal_block + j_start ;
			replay_count++ ;
		}

	if (oldest_trans_id == 0)
		{
			printk("Journal hasn't valid transactions.\n") ;
		}

	m_debug("clm-2038: Prepared %d transactions in %lu seconds, mount_id now %lu\n",
		   replay_count, CURRENT_TIME - start, j_mount_id) ;

	{  /* sorting j_table by real block number*/
		int i,k,l,n;
		int ss_sort;
		__u32 field_ex;
		k=0;
		l = j_tab_count;
		n=(l+k+1)/2;
		ss_sort = 1;
		while( ss_sort || n>1)
			{
			printk(".");
			m_debug("\nS");
				ss_sort = 0;
				for(i=0;(i+n)<l;i++)
					{
						if ((j_table+i)->r_BlockNr>(j_table+i+n)->r_BlockNr)
							{
								m_debug("s");
								field_ex                 = (j_table+i)->r_BlockNr;
								(j_table+i)->r_BlockNr   = (j_table+i+n)->r_BlockNr;
								(j_table+i+n)->r_BlockNr = field_ex;
								field_ex                 = (j_table+i)->j_BlockNr;
								(j_table+i)->j_BlockNr   = (j_table+i+n)->j_BlockNr;
								(j_table+i+n)->j_BlockNr = field_ex;
								ss_sort = 1;
							}
					}
				n=(n+1)/2;
			}
		for(i=0;i<l;i++)
			{
				m_debug("\n%3d,(%5d/%5d)",i, (j_table+i)->r_BlockNr,	(j_table+i)->j_BlockNr);
			}
	}

	printk("\n");
	kfree(desc);
	kfree(commit);
	kfree(jh);

	
	return 0 ;
}


static inline int  m__comp_keys (struct m_key * k1, struct m_key * k2, int key_size)
{
  __u32 * p_s_key1, * p_s_key2;

  p_s_key1 = (__u32 *)k1;
  p_s_key2 = (__u32 *)k2;

  if ( *p_s_key1 < *p_s_key2 )  return -1;
  if ( *p_s_key1 > *p_s_key2 )  return 1;

  ++p_s_key1;
  ++p_s_key2;

  if ( *p_s_key1 < *p_s_key2 )  return -1;
  if ( *p_s_key1 > *p_s_key2 )  return 1;
  
  if (key_size)
	  {
		  if (m_le_key_version (k1)== ITEM_VERSION_1)
			  {
				  if (k1->u.k_offset_v1.k_offset < k2->u.k_offset_v2.k_offset) return -1;
				  if (k1->u.k_offset_v1.k_offset > k2->u.k_offset_v2.k_offset) return  1;
				  if (uniqueness2type(k1->u.k_offset_v1.k_uniqueness) < k2->u.k_offset_v2.k_type)   return -1;
				  if (uniqueness2type(k1->u.k_offset_v1.k_uniqueness) > k2->u.k_offset_v2.k_type)   return  1;
			  }
		  else
			  {
				  if (k1->u.k_offset_v2.k_offset < k2->u.k_offset_v2.k_offset) return -1;
				  if (k1->u.k_offset_v2.k_offset > k2->u.k_offset_v2.k_offset) return  1;
				  if (k1->u.k_offset_v2.k_type   < k2->u.k_offset_v2.k_type)   return -1;
				  if (k1->u.k_offset_v2.k_type   > k2->u.k_offset_v2.k_type)   return  1;
			  }
	  }
  return 0;
}


#define MAX_GENERATION_NUMBER  127
#define  MAX_GEN_NUMBER 127


#if 0

#include <asm/types.h>

#define DELTA 0x9E3779B9
#define FULLROUNDS 10		/* 32 is overkill, 16 is strong crypto */
#define PARTROUNDS 6		/* 6 gets complete mixing */

#ifndef __KERNEL__
typedef __u32 u32;
#endif

/* a, b, c, d - data; h0, h1 - accumulated hash */
#define TEACORE(rounds)							\
	do {								\
		u32 sum = 0;						\
		int n = rounds;						\
		u32 b0, b1;						\
									\
		b0 = h0;						\
		b1 = h1;						\
									\
		do							\
		{							\
			sum += DELTA;					\
			b0 += ((b1 << 4)+a) ^ (b1+sum) ^ ((b1 >> 5)+b);	\
			b1 += ((b0 << 4)+c) ^ (b0+sum) ^ ((b0 >> 5)+d);	\
		} while(--n);						\
									\
		h0 += b0;						\
		h1 += b1;						\
	} while(0)


static u32 keyed_hash(const char *msg, int len)
{
	u32 k[] = { 0x9464a485, 0x542e1a94, 0x3e846bff, 0xb75bcfc3}; 

	u32 h0 = k[0], h1 = k[1];
	u32 a, b, c, d;
	u32 pad;
	int i;
 

	//	assert(len >= 0 && len < 256);

	pad = (u32)len | ((u32)len << 8);
	pad |= pad << 16;

	while(len >= 16)
		{
			a = (u32)msg[ 0]      |
				(u32)msg[ 1] << 8 |
				(u32)msg[ 2] << 16|
				(u32)msg[ 3] << 24;
			b = (u32)msg[ 4]      |
				(u32)msg[ 5] << 8 |
				(u32)msg[ 6] << 16|
				(u32)msg[ 7] << 24;
			c = (u32)msg[ 8]      |
				(u32)msg[ 9] << 8 |
				(u32)msg[10] << 16|
				(u32)msg[11] << 24;
			d = (u32)msg[12]      |
				(u32)msg[13] << 8 |
				(u32)msg[14] << 16|
				(u32)msg[15] << 24;
		
			TEACORE(PARTROUNDS);

			len -= 16;
		msg += 16;
		}
	
	if (len >= 12)
		{
	    	//assert(len < 16);
			if (len >= 16)
				*(int *)0 = 0;

			a = (u32)msg[ 0]      |
				(u32)msg[ 1] << 8 |
				(u32)msg[ 2] << 16|
				(u32)msg[ 3] << 24;
			b = (u32)msg[ 4]      |
				(u32)msg[ 5] << 8 |
				(u32)msg[ 6] << 16|
				(u32)msg[ 7] << 24;
			c = (u32)msg[ 8]      |
				(u32)msg[ 9] << 8 |
				(u32)msg[10] << 16|
				(u32)msg[11] << 24;
			
			d = pad;
			for(i = 12; i < len; i++)
				{
					d <<= 8;
					d |= msg[i];
				}
		}
	else
		if (len >= 8)
			{
				//assert(len < 12);
				if (len >= 12)
					*(int *)0 = 0;
				a = (u32)msg[ 0]      |
					(u32)msg[ 1] << 8 |
					(u32)msg[ 2] << 16|
					(u32)msg[ 3] << 24;
				b = (u32)msg[ 4]      |
					(u32)msg[ 5] << 8 |
					(u32)msg[ 6] << 16|
					(u32)msg[ 7] << 24;

				c = d = pad;
				for(i = 8; i < len; i++)
					{
						c <<= 8;
						c |= msg[i];
					}
			}
		else
			if (len >= 4)
				{
					//assert(len < 8);
					if (len >= 8)
						*(int *)0 = 0;
					a = (u32)msg[ 0]      |
						(u32)msg[ 1] << 8 |
						(u32)msg[ 2] << 16|
						(u32)msg[ 3] << 24;
					
					b = c = d = pad;
					for(i = 4; i < len; i++)
						{
							b <<= 8;
							b |= msg[i];
						}
				}
			else
				{
					//assert(len < 4);
					if (len >= 4)
						*(int *)0 = 0;
					a = b = c = d = pad;
					for(i = 0; i < len; i++)
						{
							a <<= 8;
							a |= msg[i];
						}
				}
	
	TEACORE(FULLROUNDS);
	
/*	return 0;*/
	return h0^h1;
}


static u32 yura_hash (const char *msg, int len)
{
    int j, pow;
    u32 a, c;
    int i;
    
    for (pow=1,i=1; i < len; i++) pow = pow * 10; 
    
    if (len == 1) 
		a = msg[0]-48;
    else
		a = (msg[0] - 48) * pow;
    
    for (i=1; i < len; i++)
		{
			c = msg[i] - 48; 
			for (pow=1,j=i; j < len-1; j++) pow = pow * 10; 
			a = a + c * pow;
		}
    
    for (; i < 40; i++)
		{
			c = '0' - 48; 
			for (pow=1,j=i; j < len-1; j++) pow = pow * 10; 
			a = a + c * pow;
		}
    
    for (; i < 256; i++)
		{
			c = i; 
			for (pow=1,j=i; j < len-1; j++) pow = pow * 10; 
			a = a + c * pow;
		}
    
    a = a << 7;
    return a;
}

static u32 r5_hash (const char *msg, int len)
{
  u32 a=0;
  int i;

  for (i = 0; i < len; i ++)
	  {
		  a += msg[i] << 4;
		  a += msg[i] >> 4;
		  a *= 11;
	  }
  return a;
}

#endif

static struct  hash_function_t
{
	int  hash_function_code;
	hashf_t hash_function;
	char hash_function_name[10];
}  hash_function_table[]=
{
	TEA_HASH,   keyed_hash, "tea",
	YURA_HASH,  yura_hash,  "rupasov",
	R5_HASH,    r5_hash,     "r5",
};



static hashf_t hash_function (__u32 tt)
{
	int ret;
	ret=1;
	switch (tt)
		{
		case TEA_HASH:
		case YURA_HASH:
		case R5_HASH:
			ret=tt-1;
		};
	return hash_function_table[ret].hash_function;
}


/* for debug */
#ifdef MM_DEBUG_KEY
#define vv_print_key(a,b)      vv_prn_key(0,(a),(b))
#define vv_print_key_stop(a,b) vv_prn_key(1,(a),(b))
#else
#define vv_print_key(a...)
#define vv_print_key_stop(a...) 
#endif


static inline int vv_prn_key(int opt,char *mes,struct m_key * k)
{
	printk("    %s  key= %u/%u/%lu/%u\n?",
		   mes,
		   k->k_dir_id,
		   k->k_objectid,
		   (long)k->u.k_offset_v2.k_offset,
		   (unsigned int) k->u.k_offset_v2.k_type);
	if (opt) kbd_getc();
	return;
		
}




/*make milo_item - copy unaligned space to aligned struct */
static struct milo_item * vv_m_ihead(struct milo_item * p_item, struct m_block_head * p_bh, int p_pos)
{
	int type;
#if 0
	struct
	{
		__u16 sd_mode;
		__u16 sd_reserved;
		__u32 sd_nlink;
		__u64 sd_size;	/* file size */
		__u32 sd_uid;		/* owner */
		__u32 sd_gid;		/* group */
		__u32 sd_atime;	/* time of last access */
		__u32 sd_mtime;	/* time file was last modified  */
		__u32 sd_ctime;	/* time inode (stat data) was last changed (except changes to sd_atime and sd_mtime) */
		__u32 sd_blocks;
		__u32 sd_rdev;
	}__attribute__ ((__packed__)) * sd;
	struct
	{
		__u32 deh_offset;
		__u32 deh_dir_id;
		__u32 deh_objectid;
		__u16 deh_location;
		__u16 deh_state;
	}__attribute__ ((__packed__))  * de;
#endif
	memcpy( p_item, (char *)(p_bh)+M_BLKH_SIZE+M_IH_SIZE*(p_pos), M_IH_SIZE );
			
	memcpy( &(p_item->ih_r_key), &(p_bh->blk_right_delim_key), M_KEY_SIZE);

	memcpy( &p_item->body, (char*)(p_bh)+ p_item->ih_item_location, p_item->ih_item_len );


#if 0
	sd=de=(char*)(p_bh)+ p_item->ih_item_location;
	switch (p_item->ih_key.u.k_offset_v2.k_type)
		{
		case 0:          /*stat data*/
			if ((p_item->body.sd.sd_mode!=sd->sd_mode) ||
				(p_item->body.sd.sd_size!=sd->sd_size) ||
				(p_item->body.sd.sd_uid!=sd->sd_uid) ||
				(p_item->body.sd.sd_gid!=sd->sd_gid) ||
				(p_item->body.sd.sd_atime!=sd->sd_atime) ||
				(p_item->body.sd.sd_mtime!=sd->sd_mtime) ||
				(p_item->body.sd.sd_ctime!=sd->sd_ctime) )
				{
					printk("\nstatdata unaligned trap\n  %p=%d/%ld/%d/%d/%d/%d/%d  %p=%d/%ld/%d/%d/%d/%d/%d\n?",
						   p_item,
						   p_item->body.sd.sd_mode,
						   p_item->body.sd.sd_size,
						   p_item->body.sd.sd_uid,
						   p_item->body.sd.sd_gid,
						   p_item->body.sd.sd_atime,
						   p_item->body.sd.sd_mtime,
						   p_item->body.sd.sd_ctime,
						   sd,
						   sd->sd_mode,
						   sd->sd_size,
						   sd->sd_uid,
						   sd->sd_gid,
						   sd->sd_atime,
						   sd->sd_mtime,
						   sd->sd_ctime);
					kbd_getc();
				}
			break;
		case 1:
			break;
		case 2:
			break;
		case 3:
			{
				int i;
				for ( i=0; i<p_item->u.ih_entry_count ; i++,de++ )
					if ( p_item->body.deh[i].deh_offset != de->deh_offset ||
						 p_item->body.deh[i].deh_dir_id != de->deh_dir_id ||
						 p_item->body.deh[i].deh_objectid != de->deh_objectid ||
						 p_item->body.deh[i].deh_location != de->deh_location )
						{
							printk("\ndir entry unaligned trap\n  %d/%d/%d/%d  %p=%d/%d/%d/%d\n?",
								   p_item->body.deh[i].deh_offset ,
								   p_item->body.deh[i].deh_dir_id ,
								   p_item->body.deh[i].deh_objectid ,
								   p_item->body.deh[i].deh_location ,
								   de,
								   de->deh_offset ,
								   de->deh_dir_id,
								   de->deh_objectid,
								   de->deh_location);
							kbd_getc();
						}
			}
			
			break;
		default:
			printk("\n error unaligned trap type =%d\n?",p_item->ih_key.u.k_offset_v2.k_type);
					kbd_getc();
			
		}
#endif

	if (m_le_key_version (p_item)== ITEM_VERSION_1 && p_item->ih_key.u.k_offset_v1.k_uniqueness)
		{
			type                                   = uniqueness2type(p_item->ih_key.u.k_offset_v1.k_uniqueness);
			p_item->ih_key.u.k_offset_v2.k_offset  = p_item->ih_key.u.k_offset_v1.k_offset;
			p_item->ih_key.u.k_offset_v2.k_type    = type;
		}


	
	p_item->it_Number = p_pos;
	p_item->it_Last = (p_pos == (p_bh->blk_nr_item-1)) ;
	vv_print_key("get item head",p_item);
	return p_item;
}


static int reiserfs_m_mount(long device, int quiet)
{
	extern void set_blocksize(kdev_t dev, int size);
	int i;
	long sb_offset;
     /* Initialize the inode table */
	for (i = 0; i < MAX_OPEN_FILES; i++)
		{
			inode_table[i].free = 1;
			inode_table[i].inumber = 0;
		}
	
	set_blocksize(device, REISERFS_BLOCK_SIZE);
	sb_offset = REISERFS_DISK_OFFSET_IN_BYTES;
	if (device_read(fd, (char *) &sb, sizeof(sb), sb_offset) !=  sizeof(sb))
		{
			printk("reiserfs sb read failed");
			device_close(fd);
			return (-1);
		}
	if (!is_reiserfs_magic_string(&sb))
		{
			printk ("reiserfs_read_super: can't find a reiserfs filesystem.\n");
			device_close(fd);		
			return -1;
		}
	blocksize = REISERFS_BLOCK_SIZE;
	sb.s_blocksize = REISERFS_BLOCK_SIZE;
	reiserfs.blocksize = REISERFS_BLOCK_SIZE;
	p_hash_function = hash_function (sb.s_hash_function_code);
	s_dev=device;
	j_tab_count=0;
	if (m_journal_read(&sb) < 0)
		{
			printk("\nJournal failure. Unable replay");
			j_tab_count=0;
		}
	return (0);
}


static long j_device_read(int fd,char * p_buf, int blocksize, __u32 blkno)
{
	unsigned long offset;

	offset = blocksize *  (j_tab_count ? check_j_block(blkno) : blkno);
	return ( device_read( fd, p_buf, blocksize, offset) );
}


static struct milo_item * reiserfs_search_item(struct m_key * p_key, char * p_buf, struct milo_item *p_item)
{
	int i, j, l,h,r,s;
	unsigned long offset;
	struct m_key right_delim_key={-1,-1,-1L, 1};

	vv_print_key("search item",p_key);	

	if ( m_comp_keys((struct m_key *) p_item, p_key ))
		{
	
			bh = ( struct m_block_head *) p_buf;

			if ( j_device_read( fd, p_buf, blocksize, sb.s_root_block) !=blocksize )
				{
					printk("reiserfs_search_item(read root of tree): read error\n");
					return (NULL);
				}

			while ( bh->blk_level > 1 )  /* Search leaf*/
				{
					l = 0;
					h = VV_ITEMS(bh)-1;
					i = ( l + h )/2;
					s=0;
					while( l < h )
						{
							m_debug("\n %d? r=%d:(%d<%d<%d)",s,r,l,i,h);
							r = m_comp_keys( VV_KEY(bh,i), p_key );
							if ( !r    ) break;
							if ( s )
								{
									if ( r >= 0 || s++>1 ) break;
									i++;
								}
							else
								{
									if ( r < 0 ) l = i; 
									if ( r > 0 ) h = i;
									if ( (l+1) == h ) { i = l; s = 1;}
									else                i = ( l + h )/2;
								}
						}
					vv_print_key("inner block ",VV_KEY(bh,i));
					if ( r > 0 )
						{
							m1_copy_key( &right_delim_key, VV_KEY(bh,i));
						}
					else
						{
							i++;
							if ((i+1) < VV_ITEMS(bh)) m1_copy_key( &right_delim_key, VV_KEY(bh,i+1));
						}
					m_debug("prt#%d,%u, ", i, VV_CHILD_NUM(bh,i));
					if ( j_device_read( fd, p_buf, blocksize, VV_CHILD_NUM(bh,i) ) !=blocksize )
						{
							printk("reiserfs_search_item(next read tree): read error\n");
							return (NULL);
						}
				};
			j = 0;
			vv_print_key("right_delim_key",&right_delim_key);
			m_copy_key(&(bh-> blk_right_delim_key), &right_delim_key);
	
			/* search item */
			l = 0;
			h = VV_ITEMS(bh) - 1;
			j=0;
			r=-1;
			s=0;
			while(l<h)
				{
					m_debug("\n %d? r=%d:(%d<%d<%d)",s,r,l,j,h);
					if ( !r    ) break ;
					if (s)
						{
							if ( r <=  0  || s++ > 1) break;
							j--;
						}
					else
						{
							if ( r < 0 ) l = j; 
							if ( r > 0 ) h = j; 
							if ( ( l + 1 ) == h ) { j = h; s = 1; }
							else                    j = ( l + h )/2;
						}
					r =  m_comp_keys( (struct m_key *) vv_m_ihead( p_item, bh, j), p_key );
				}
		}
	vv_print_key_stop("founded key",p_item);
    return (p_item);	
}


static struct reiserfs_m_inode *reiserfs_m_iget( struct m_key * ino_k, struct milo_item * tmp_item)
{
    int i;
	struct reiserfs_m_inode *ip;
	struct inode_table_entry *itp;

	ip = NULL;
	itp = NULL;
    for (i = 0; i < MAX_OPEN_FILES; i++)
		{
			if (inode_table[i].free || inode_table[i].inumber==ino_k->k_objectid)
				{
					itp = &(inode_table[i]);
					ip = &(itp->inode);
					break;
				}
		}
	if ((ip == NULL) || (itp == NULL))
		{
			printk("reiserfs_m_iget: no free inodes\n");
			return (NULL);
		}
	
	itp->inumber = ino_k->k_objectid;
	itp->free = 0;
		
    ip->i_mode = 0;

    ip->i_key.k_dir_id = ino_k->k_dir_id;
	ip->i_key.k_objectid = ino_k->k_objectid;
    ip->i_key.u.k_offset_v2.k_offset = SD_OFFSET;
    ip->i_key.u.k_offset_v2.k_type = TYPE_STAT_DATA;

	reiserfs_search_item( (struct m_key *)ip, &iblkbuf , tmp_item);

	if (tmp_item->ih_version)
		{
			ip->i_mode               = le16_to_cpu (tmp_item->body.sd.sd_mode);
			ip->i_uid                = le32_to_cpu (tmp_item->body.sd.sd_uid);
			ip->i_gid                = le32_to_cpu (tmp_item->body.sd.sd_gid);
			ip->i_size               = le64_to_cpu (tmp_item->body.sd.sd_size);
			ip->i_mtime              = le32_to_cpu (tmp_item->body.sd.sd_mtime);
			ip->i_atime              = le32_to_cpu (tmp_item->body.sd.sd_atime);
			ip->i_ctime              = le32_to_cpu (tmp_item->body.sd.sd_ctime);
			ip->i_blocks             = le32_to_cpu (tmp_item->body.sd.sd_blocks);
		}
	else
		{
			ip->i_mode               = le16_to_cpu (tmp_item->body.sd1.sd_mode);
			ip->i_uid                = le16_to_cpu (tmp_item->body.sd1.sd_uid);
			ip->i_gid                = le16_to_cpu (tmp_item->body.sd1.sd_gid);
			ip->i_size               = le32_to_cpu (tmp_item->body.sd1.sd_size);
			ip->i_mtime              = le32_to_cpu (tmp_item->body.sd1.sd_mtime);
			ip->i_atime              = le32_to_cpu (tmp_item->body.sd1.sd_atime);
			ip->i_ctime              = le32_to_cpu (tmp_item->body.sd1.sd_ctime);
			ip->i_blocks             = le32_to_cpu (tmp_item->body.sd1.sd_blocks);
		}


	switch ( ip->i_mode & S_IFMT )
		{
		case S_IFREG:
			ip->i_key.u.k_offset_v2.k_offset = 1;
			ip->i_key.u.k_offset_v2.k_type   = TYPE_DIRECT;
			reiserfs_search_item( (struct m_key *)ip, &iblkbuf , tmp_item);
//			memcpy(&(ip->i_data),&(tmp_item->body),tmp_item->ih_item_len);
//			ip->i_data_length = tmp_item->ih_item_len/sizeof(__u32);
			break;
		case S_IFDIR:
			ip->i_key.u.k_offset_v2.k_offset = DOT_OFFSET;
			ip->i_key.u.k_offset_v2.k_type = TYPE_DIRENTRY;
			reiserfs_search_item( (struct m_key *)ip, &iblkbuf , tmp_item);
/*
//			memcpy(&(ip->i_data),&(tmp_item->body),tmp_item->ih_item_len);
//			ip->i_data_length = tmp_item->ih_item_len;
*/			
			break;
		case S_IFCHR:
			break;
		case S_IFBLK:
		case S_IFIFO:
		case S_IFSOCK:
			break;
		case S_IFLNK:
			(ip->i_key.u.k_offset_v2.k_offset)= 1;
			ip->i_key.u.k_offset_v2.k_type = TYPE_DIRECT;
			reiserfs_search_item( (struct m_key *)ip, &iblkbuf , tmp_item);
/*
//			memcpy(&(ip->i_data),&(tmp_item->body),tmp_item->ih_item_len);
//			ip->i_data_length = tmp_item->ih_item_len;
*/
			break;
		}
	return (ip);
}

/* Release our hold on an inode.  Since this is a read-only application,
 * don't worry about putting back any changes...
 */
static void reiserfs_m_iput(struct reiserfs_m_inode *ip)
{
	struct inode_table_entry *itp;

	/* Find and free the inode table slot we used... */
	itp = (struct inode_table_entry *) ip;

	itp->inumber = 0;
	itp->free = 1;
}

/* Return the size of the open file */
static long reiserfs_m_fdsize(int fd)
{
	struct reiserfs_m_inode *ip;

	ip = &(inode_table[fd].inode);
	return ip->i_size;
}

/* Read block number "blkno" from the specified file */
static int reiserfs_m_bread(int fd, long blkno, char *buffer)
{
	struct reiserfs_m_inode *ip;
	static struct inode_table_entry * itp;
	ip = &(inode_table[fd].inode);
	itp = ip;
	if (itp->free)
		{
			printk("file %d not open", fd);
			return -1;
		}
	reiserfs_m_breadi(ip, blkno, buffer);
	return 0;
}

#ifndef IOBUF_SIZE
#define IOBUF_SIZE 1024
#endif

static void reiserfs_m_breadi(struct reiserfs_m_inode *ip, long p_blkno, char *buffer)
{
	long dev_blkno,blkno;
	struct m_key pos_key;
	unsigned long offset;
	unsigned long offset_min;
	unsigned long offset_max;
	unsigned int	n_pos;
	unsigned long r_size;

	offset = p_blkno * IOBUF_SIZE   /*1024*/  ;

	if (offset > ip->i_size)
		{
			printk("\nattempt read %d belong end of file:%d",offset,ip->i_size);
			memset(buffer,0,blocksize);
			kbd_getc();
			return;
		}
	
	offset_min = it_open.ih_key.u.k_offset_v2.k_offset-1;
	offset_max = it_open.ih_item_len-1;

	if(it_open.ih_key.u.k_offset_v2.k_type==TYPE_INDIRECT )
			offset_max = offset_max * blocksize + offset_min;

	if ( offset < offset_min || offset > offset_max )
		{
			m_copy_key (&pos_key, ip);
			pos_key.u.k_offset_v2.k_offset = offset;
			reiserfs_search_item( &pos_key, buffer, &it_open);
			vv_print_key(" bread: find new item",&pos_key);
		}

	switch(it_open.ih_key.u.k_offset_v2.k_type)
		{
		case TYPE_INDIRECT :

			offset -= ( it_open.ih_key.u.k_offset_v2.k_offset - 1 );

			n_pos  = offset / blocksize;
			offset = offset % blocksize;
			dev_blkno = it_open.body.item[n_pos];
			m_debug("%d",dev_blkno);
			
			if (j_tab_count) dev_blkno=check_j_block(dev_blkno);

			offset += dev_blkno * blocksize;
			
			m_debug("%d",dev_blkno);

			if ( device_read (fd, buffer,   IOBUF_SIZE ,  offset) != IOBUF_SIZE )
				{
					printk("reiserfs_bread: read error");
				}
			break;
		case /*V1_DIRECT_UNIQUENESS*/ TYPE_DIRECT :
			memcpy( buffer, it_open.body.txt, it_open.ih_item_len );
			break;
		}
	return 0;
}


static __u32 m_get_third_component ( const char * name, int len)
{
	__u32 res;
	if (!len || (len == 1 && name[0] == '.'))           return DOT_OFFSET;
	if (len == 2 && name[0] == '.' && name[1] == '.')   return DOT_DOT_OFFSET;
	res = p_hash_function (name, len);
	res = M_GET_HASH_VALUE(res);
	if (res == 0) 	res = 128;
	return res +  MAX_GENERATION_NUMBER /*MAX_GEN_NUMBER*/;
}


static struct reiserfs_m_inode *reiserfs_m_namei(char *name, struct milo_item *n_item)
{
	char namebuf[512];
	char *component;
	struct reiserfs_m_inode
		*dir_inode = NULL,
		*file_inode;
	int i;
	int prev_offset;
struct reiserfs_m_de_head * de;
	

	struct m_key pos_key;	/* key of current position in the directory (key of directory entry) */
    struct m_key * rkey;
	int link_count;
	
	strcpy(namebuf, name);
	
  abs:
	file_inode =  reiserfs_m_iget ( &m_root_key, n_item);

  rel:
	if (!file_inode)
		return NULL;

	component = strtok(namebuf, "/");
	while (component)
		{
			int component_length;

			if (dir_inode)  reiserfs_m_iput(dir_inode);
			dir_inode = file_inode;
			file_inode = NULL;
			
			component_length = strlen(component);
			
#ifdef MILO_SYMLINK_DIRS
			if (S_ISLNK(dir_inode->i_mode))
				{
					m_debug("     symlink     ");
					if (++link_count > 10)
						printk("%s: possibly circular symlink\n", name);
					else
						{
							char *t;
							reiserfs_search_item( dir_inode, &iblkbuf, n_item);
							memcpy( blkbuf, n_item->body.txt, n_item->ih_item_len );
							t = blkbuf + strlen(blkbuf);
							while (component != NULL)
								{
									t = stpcpy(stpcpy(t, "/"), component);
									component = strtok(NULL, "/");
								}

							/* Put it back to namebuf */
							strcpy(namebuf, blkbuf);

							if (*namebuf == '/')
								{
									/* Absolute symlink */
									if (dir_inode)
										reiserfs_m_iput(dir_inode);
									if (file_inode)
										reiserfs_m_iput(file_inode);
									goto abs;
								}
							else
								{
									/* Relative symlink */
									if (file_inode)
										reiserfs_m_iput(file_inode);
									file_inode = dir_inode;
									goto rel;
								}
						}
				}
#endif				/* MILO_SYMLINK_DIRS */
			
			if (!S_ISDIR(dir_inode->i_mode))
				{
					printk("%s: path component is not a directory\n", name);
					break;
				}

			m_copy_key (&pos_key, dir_inode);
			pos_key.u.k_offset_v2.k_offset = m_get_third_component( component, component_length);

			reiserfs_search_item( (struct m_key *)dir_inode, &iblkbuf , n_item);

			de = n_item->body.deh;
			
			i = n_item->u.ih_entry_count;

			prev_offset = n_item->ih_item_len;
			
			while ( i  && M_GET_HASH_VALUE( de->deh_offset ) < M_GET_HASH_VALUE( pos_key.u.k_offset_v2.k_offset ) )
				{
					prev_offset = de->deh_location;
					de++;
					i--;
				}

			if ( strncmp( n_item->body.txt + de->deh_location , component, prev_offset - de->deh_location ) )
				{
					reiserfs_m_iput(dir_inode);
					return NULL;
				}
			pos_key.k_dir_id    = de->deh_dir_id;
			pos_key.k_objectid  = de->deh_objectid;
			pos_key.u.k_offset_v2.k_offset    = DOT_OFFSET;
			pos_key.u.k_offset_v2.k_type      = TYPE_DIRENTRY;

			file_inode = reiserfs_m_iget(&pos_key,n_item);
			
			component = strtok(NULL, "/");
		}

	if (dir_inode)
		reiserfs_m_iput(dir_inode);
	return (file_inode);
}

/*
** read next byte
*/

static char vv_next(long count, struct reiserfs_m_inode *ip,char * blkh, long f_size)
{
	unsigned int blkno;
	unsigned int offs;
	
	if ( count < f_size )
		{
			blkno = count / IOBUF_SIZE;
			offs  = count % IOBUF_SIZE;
			if ( !offs )
					reiserfs_m_breadi( ip, blkno, blkh);
			return (*(blkh+offs));
		}
	else return 0;
}


/*
** print file  
*/
static void reiserfs_m_cat(struct reiserfs_m_inode *ip, char * buf)
	/*char *filename, char * dev)*/
{
	long f_size;
	__u32 blkno;
	int fd;
	int i,j;
	char * blkh;
	char * component;
	struct inode_table_entry * itp;
	char cc;
	reiserfs_search_item( ip, buf, &it_open);

	blkh = buf;

	itp = (struct inode_table_entry *) ip;
	fd = (itp - inode_table);

	f_size=reiserfs_m_fdsize(fd);
	printk( "file size=%d\n", f_size);
	blkno=0;
	i=0;
	j=0;
	printk("\n%3d:",++i);
	while( cc = vv_next( j++, ip, blkh, f_size))
		{
/*			printk( "%c",(isprint(cc))? cc:'.' );*/
			printk( "%c",(cc<12)? '.':cc );
			if ( cc == '\n' )
				{
					if( !( i%20 ) )
						if ( kbd_getc() == 'q' )
							break;
					printk( "%3d:", ++i );
				}
		}

	printk("\n");
	reiserfs_m_close(fd);
}

/*
** ls (*dirname, void * devname)
** devname - not use, no need: milo read from 1 device, opened befor
** find dirname, and list it contains.
**   if finded name is file- trie print it
*/
static void reiserfs_m_ls(char *dirname, char *devname)
{
	char filename[512];
	int filenamelen;
	char format_string[16];
	struct reiserfs_m_inode *dir_inode;
	struct reiserfs_m_inode *file_inode;
	long entr_count=0;
	int columns;
	int col;
	char typetag;
	struct m_block_head * blkh;
	struct reiserfs_m_de_head * de;

	struct m_key pos_key;	/* key of current position in the directory (key of directory entry) */
    struct m_key * rkey;
    char small_buf[REISERFS_BLOCK_SIZE] ;
	int i;
	__u32 prev_loc;

#define NOT_EOD NotEOD
	int NotEOD;



	strcpy(format_string,"   %s");
                                                /* Get the inode for the specified directory */
	dir_inode = reiserfs_m_namei(dirname,&it_buf);

	printk("\nls %s",dirname);
	if (!dir_inode)
		{
			printk("%s: is no  file or directory\n", dirname);
			return;
		}

	/* If this inode exists but is *not* a directory, just cat it here */
	if ((dir_inode->i_mode & S_IFMT) != S_IFDIR)
		{
			reiserfs_m_cat(dir_inode, &it_buf);
			return;
		}

	col = 0;
	columns = 4;
	NOT_EOD = 1;
	blkh = &small_buf;
	
	while ( NOT_EOD )
		{
			NOT_EOD = 0;

			reiserfs_search_item( dir_inode, blkh, &it_buf);
			de = it_buf.body.deh;
			i = it_buf.u.ih_entry_count;
			prev_loc = it_buf.ih_item_len;
//			de++;
//			prev_loc = (de++)->deh_location;
//			i-=2;
			printk("\n");


			while ( i  )  /* loop by dir_item*/
				{
					entr_count++;
					filenamelen          = prev_loc - de->deh_location;
					memcpy( filename, it_buf.body.txt + de->deh_location ,filenamelen);
					filename[filenamelen] = '\0';
					printk(format_string, filename);
					pos_key.k_dir_id     = de->deh_dir_id;
					pos_key.k_objectid   = de->deh_objectid;
					pos_key.u.k_offset_v2.k_offset     = 1;
					pos_key.u.k_offset_v2.k_type = TYPE_DIRENTRY;

					file_inode = reiserfs_m_iget(&pos_key,&n_item_buf);

					vv_print_key("pos_key",&pos_key);

					if (file_inode)
						{

/*
							typetag = ' ';
							if ( file_inode->i_mode & S_IFMT == S_IFDIR) typetag = '/';
							if ( file_inode->i_mode & S_IFMT == S_IFLNK) typetag = '@';
							if ( file_inode->i_mode & S_IFMT == S_IFIFO) typetag = '=';
*/							
							switch ( file_inode->i_mode & S_IFMT )
								{
								case S_IFLNK:
									typetag = '@';
									break;
								case S_IFDIR:
									typetag = '/';
									break;
								case S_IFIFO:
								case S_IFSOCK:
									typetag = '=';
									break;
								default:
									typetag = ' ';
									break;
									}
							reiserfs_m_iput(file_inode);
						}
					else
						{
							typetag = '!';
						}
					printk("%c", typetag);
					
					if ( !(++col % columns) )  printk("\n");
					
					prev_loc = de->deh_location;

					de++;
					i--;
				}

			if ( it_buf.it_Last ) /* This item is the last in block */
				{
					copy_key(&pos_key,&(it_buf.ih_r_key));
					vv_print_key    ("  tri new pos",&pos_key);
					reiserfs_search_item( &pos_key, blkh, &it_buf);
					vv_print_key_stop     ("  founded pos",&it_buf.ih_key);
					
					if (!m_comp_short_keys(&(dir_inode->i_key),&it_buf.ih_key) &&
						it_buf.ih_key.u.k_offset_v2.k_offset>m_get_third_component(filename,filenamelen) &&
						it_buf.ih_key.u.k_offset_v2.k_type==TYPE_DIRENTRY)
						{                                      	/*directory continues in next item*/
							copy_key(&(dir_inode->i_key),&it_buf.ih_key);
							NOT_EOD = 1;
						}
				}
		}			/* end-while */
	
	if ((col % columns))
		{
			printk("\n");
		}
	printk("\ntotal %d entries\n",entr_count);

	/* We're done with the directory now... */
	reiserfs_m_iput(dir_inode);

}

static int reiserfs_m_open(char *filename)
{
	/* Unix-like open routine.  Returns a small integer (actually an index into
	 * the inode table... */
	struct reiserfs_m_inode *ip;
	ip = reiserfs_m_namei(filename,&it_open);

	if (ip)
		{
			struct inode_table_entry *itp;

			reiserfs_search_item( ip, &n_item_buf, &it_open);
			m_debug("\n open: %s size:%d itemsize:%d fist block:%d",
				   filename, ip->i_size,  it_open.ih_item_len,it_open.body.item[0]);
			itp = (struct inode_table_entry *) ip;
			return (itp - inode_table);
		}
	else
		{
			printk("not opened");
			return (-1);
		}
}


static void reiserfs_m_close(int fd)
{
	reiserfs_m_iput(&(inode_table[fd].inode));
	device_close(fd);
}
